import re

__all__ = ["split_streetnr", "getter_prefixed_address", "CARD_TYPES", "get_credit_card_issuer"]

_extract_streetnr = re.compile(r"([0-9]+)$")
def split_streetnr(address, fallback=None):
    address = address.strip()
    # to prevent regex attacks, limit to 15 chars
    ret = _extract_streetnr.search(address[-15:])
    if ret:
        # use rstrip because left side is stripped already
        return address[:(ret.start()-15)].rstrip(), ret.group(0)
    else:
        return address, fallback

def getter_prefixed_address(prefix):
    """ create getter for prefixed address format """
    email = "{}_email".format(prefix)
    first_name = "{}_first_name".format(prefix)
    last_name = "{}_last_name".format(prefix)
    address_1 = "{}_address_1".format(prefix)
    address_2 = "{}_address_2".format(prefix)
    city = "{}_city".format(prefix)
    postcode = "{}_postcode".format(prefix)
    country_code = "{}_country_code".format(prefix)
    country_area = "{}_country_area".format(prefix)
    def _get_address(self):
        return {
            "email": getattr(self, email, None),
            "first_name": getattr(self, first_name, None),
            "last_name": getattr(self, last_name, None),
            "address_1": getattr(self, address_1, None),
            "address_2": getattr(self, address_2, None),
            "city": getattr(self, city, None),
            "postcode": getattr(self, postcode, None),
            "country_code": getattr(self, country_code, None),
            "country_area": getattr(self, country_area, None)
        }
    return _get_address


CARD_TYPES = [
    (re.compile(r'^4[0-9]{12}(?:[0-9]{3})?$'), 'visa', 'VISA'),
    (re.compile(r'^5[1-5][0-9]{14}$'), 'mastercard', 'MasterCard'),
    (re.compile(r'^6(?:011|5[0-9]{2})[0-9]{12}$'), 'discover', 'Discover'),
    (re.compile(r'^3[47][0-9]{13}$'), 'amex', 'American Express'),
    (re.compile(r'^(?:(?:2131|1800|35\d{3})\d{11})$'), 'jcb', 'JCB'),
    (re.compile(r'^(?:3(?:0[0-5]|[68][0-9])[0-9]{11})$'), 'diners', 'Diners Club'),
    (re.compile(r'^(?:5[0678]\d\d|6304|6390|67\d\d)\d{8,15}$'), 'maestro', 'Maestro')]


def get_credit_card_issuer(number):
    for reg, card_type, name in CARD_TYPES:
        if reg.match(number):
            return card_type, name
    return None, None


class DictInputWrapper(object):
    """
    Allows dicts to be provided as formdict
    """

    def __init__(self, _dict):
        self._wrapped = _dict

    def __iter__(self):
        return iter(self._wrapped)

    def __len__(self):
        return len(self._wrapped)

    def __contains__(self, name):
        return (name in self._wrapped)

    def getlist(self, name, default=None):
        if name in self._wrapped:
            return [self._wrapped[name]]
        else:
            return default



from unittest import TestCase
from unittest.mock import patch, NonCallableMock, MagicMock

import simplejson as json

from . import get_base_url
from .models import BasePayment
from .urls import _process_data
from .. import ProviderVariant, PROVIDER_CACHE, PaymentStatus, HttpRequest, RedirectNeeded
from web_payments_dummy import DummyProvider

class TestHelpers(TestCase):
    @patch('django.conf.settings.PAYMENT_HOST', new_callable=NonCallableMock)
    def test_text_get_base_url(self, host):
        host.__str__ = lambda x: "example.com/string"
        self.assertEqual(get_base_url(), "https://example.com/string")

    @patch('django.conf.settings.PAYMENT_HOST')
    def test_callable_get_base_url(self, host):
        host.return_value = "example.com/callable"
        self.assertEqual(get_base_url(), "https://example.com/callable")

    def test_model(self):
        BasePayment.load_providers()
        testp = BasePayment()
        for i in BasePayment.list_providers():
            self.assertIsInstance(i, ProviderVariant)
            self.assertIn(i.extra["name"], PROVIDER_CACHE)
            testp.variant = i.extra["name"]
            self.assertIs(testp.provider, PROVIDER_CACHE[i.extra["name"]])


class TestUrl(TestCase):
    @patch('web_payments.django.models.BasePayment.get_failure_url')
    @patch('web_payments.django.models.BasePayment.get_success_url')
    @patch('web_payments.django.models.BasePayment.save')
    def test_body(self, save, success, failure):
        success.return_value = "success.example.com"
        failure.return_value = "failure.example.com"
        verification_status = PaymentStatus.CONFIRMED
        request = MagicMock()
        request.GET = {}
        request.method = "POST"
        request.body = json.dumps({'verification_result': verification_status})
        request.content_type = "application/json"
        testp = BasePayment(variant="DummyProvider")
        self.assertEqual(_process_data(request, testp, testp.provider).url, testp.get_success_url())
        self.assertEqual(testp.status, verification_status)

    @patch('web_payments.django.models.BasePayment.get_failure_url')
    @patch('web_payments.django.models.BasePayment.get_success_url')
    @patch('web_payments.django.models.BasePayment.save')
    def test_www(self, save, success, failure):
        success.return_value = "success.example.com"
        failure.return_value = "failure.example.com"
        verification_status = PaymentStatus.CONFIRMED
        request = MagicMock()
        request.GET = {}
        request.method = "POST"
        request.content_type = "application/x-www-form-urlencoded"
        request.POST = {'verification_result': verification_status}
        testp = BasePayment(variant="DummyProvider")
        self.assertEqual(_process_data(request, testp, testp.provider).url, testp.get_success_url())
        self.assertEqual(testp.status, verification_status)

    @patch('web_payments.django.models.BasePayment.get_failure_url')
    @patch('web_payments.django.models.BasePayment.get_success_url')
    @patch('web_payments.django.models.BasePayment.save')
    def test_get(self, save, success, failure):
        success.return_value = "success.example.com"
        failure.return_value = "failure.example.com"
        verification_status = PaymentStatus.CONFIRMED
        request = MagicMock()
        request.GET = {'verification_result': verification_status}
        request.POST = {}
        request.method = "GET"
        request.content_type = None
        testp = BasePayment(variant="DummyProvider")
        self.assertEqual(_process_data(request, testp, testp.provider).url, testp.get_success_url())
        self.assertEqual(testp.status, verification_status)




from wtforms import SelectField, ValidationError, validators

from web_payments.forms import PaymentForm
from web_payments import FraudStatus, PaymentStatus

__all__ = ["DummyForm"]

class DummyForm(PaymentForm):
    RESPONSE_CHOICES = (
        ('3ds-disabled', '3DS disabled'),
        ('3ds-redirect', '3DS redirect'),
        ('failure', 'Gateway connection error'),
        ('payment-error', 'Gateway returned unsupported response')
    )
    status = SelectField(choices=PaymentStatus.CHOICES, validators=[validators.InputRequired()])
    fraud_status = SelectField(choices=FraudStatus.CHOICES, validators=[validators.InputRequired()])
    gateway_response = SelectField(choices=RESPONSE_CHOICES, validators=[validators.InputRequired()])
    verification_result = SelectField(choices=PaymentStatus.CHOICES+[("", "")])

    def validate(self):
        if self.gateway_response.data == '3ds-redirect' and not self.verification_result.data:
            self.errors["gateway_response"] = ['When 3DS is enabled you must set post validation status']
        return super().validate()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-04-05 21:07
from __future__ import unicode_literals

from decimal import Decimal
from django.db import migrations, models
import web_payments.logic
import web_payments.translation


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='QPayment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('variant', models.CharField(max_length=255)),
                ('status', models.CharField(choices=[('waiting', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Waiting for confirmation', None)), ('preauth', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Pre-authorized', None)), ('confirmed', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Confirmed', None)), ('rejected', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Rejected', None)), ('refunded', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Refunded', None)), ('error', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Error', None)), ('input', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Input', None))], default='waiting', max_length=10)),
                ('message', models.TextField(blank=True, default='')),
                ('fraud_status', models.CharField(choices=[('unknown', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Unknown', None)), ('accept', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Passed', None)), ('reject', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Rejected', None)), ('review', web_payments.translation._lazy_constant(('web_payments.translation.translation', 'gettext'), 'Review', None))], default='unknown', max_length=10)),
                ('fraud_message', models.TextField(blank=True, default='')),
                ('extra_data', models.TextField(blank=True, default='')),
                ('token', models.CharField(blank=True, default='', max_length=36)),
                ('transaction_id', models.CharField(blank=True, max_length=255)),
                ('currency', models.CharField(max_length=10)),
                ('total', models.DecimalField(decimal_places=8, default=Decimal('0.0'), max_digits=20)),
                ('captured_amount', models.DecimalField(decimal_places=8, default=Decimal('0.0'), max_digits=20)),
                ('billing_first_name', models.CharField(blank=True, max_length=256)),
                ('billing_last_name', models.CharField(blank=True, max_length=256)),
                ('billing_address_1', models.CharField(blank=True, max_length=256)),
                ('billing_address_2', models.CharField(blank=True, max_length=256)),
                ('billing_email', models.EmailField(blank=True, max_length=254)),
                ('billing_city', models.CharField(blank=True, max_length=256)),
                ('billing_postcode', models.CharField(blank=True, max_length=20)),
                ('billing_country_code', models.CharField(blank=True, max_length=2)),
                ('billing_country_area', models.CharField(blank=True, max_length=256)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, web_payments.logic.BasicPayment),
        ),
    ]



# better include from web_payments

__all__ = []

class RedirectNeeded(Exception):
    pass

class ExternalPostNeeded(Exception):
    pass

class PaymentError(Exception):
    def __init__(self, message, code=None, gateway_message=None):
        super().__init__(message)
        self.code = code
        self.gateway_message = gateway_message

class NotInitialized(NotImplementedError):
    pass

class NotSupported(Exception):
    pass



from urllib.parse import urljoin
from web_payments.django import get_base_url
from web_payments.django.models import BasePaymentWithAddress
from django.urls import reverse

class QPayment(BasePaymentWithAddress):

    def get_success_url(self):
        return urljoin(get_base_url(), reverse("select-form"))

    def get_failure_url(self):
        return urljoin(get_base_url(), reverse("select-form"))



from decimal import Decimal

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.urls import reverse_lazy, reverse
from django.http import HttpResponseRedirect
from django.views.generic import FormView
from wtforms import SelectField, validators, StringField, DecimalField, widgets
from web_payments.forms import PaymentForm
from web_payments.django import get_payment_model
from web_payments import RedirectNeeded

class PaymentView(SuccessMessageMixin, FormView):
    template_name = "form.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['safe_urls'] = ["", reverse("payment-form"), reverse("select-form")]
        context["mytitle"] = "Payment"
        return context

    def get_form(self, form_class=None):
        payment = get_payment_model().objects.get(id=self.request.session["paymentid"])
        return payment.get_form(self.get_form_kwargs().get("data", None))

    def get(self, request, *args, **kwargs):
        try:
            return super().get(request, *args, **kwargs)
        except RedirectNeeded as exc:
            messages.add_message(request, messages.SUCCESS, "Payment redirects to %s" % exc.args[0])
            return HttpResponseRedirect(exc.args[0])

    def post(self, request, *args, **kwargs):
        """
        Handles POST requests, instantiating a form instance with the passed
        POST variables and then checked for validity.
        """
        try:
            form = self.get_form()
        except RedirectNeeded as exc:
            messages.add_message(request, messages.SUCCESS, "Payment redirects to %s" % exc.args[0])
            return HttpResponseRedirect(exc.args[0])
        #except Exception as exc:
        #    return HttpResponseBadRequest(exc, content_type="text/plain")

        if form.validate():
            messages.add_message(request, messages.SUCCESS, "Payment succeeded")
            return self.form_valid(form)
        else:
            messages.add_message(request, messages.ERROR, "Payment failed")
            return self.form_invalid(form)

class SelectPaymentForm(PaymentForm):
    variant = SelectField("Payment Method", validators=[validators.InputRequired()])
    total = DecimalField("Total amount", validators=[])
    currency = StringField("Currency", validators=[])
    billing_first_name = StringField("First Name", validators=[validators.Length(max=255)])
    billing_last_name = StringField("Last Name", validators=[validators.Length(max=255)])
    billing_address_1 = StringField("Address", validators=[validators.Length(max=255)])
    billing_address_2 = StringField("Address extension", validators=[validators.Length(max=255)])
    billing_email = StringField("Email", widget=widgets.Input("email"), validators=[validators.Length(max=255), validators.Email(), validators.Optional()])
    billing_city = StringField("City", validators=[validators.Length(max=255)])
    billing_postcode = StringField("Post code", validators=[validators.Length(max=20)])
    billing_country_code = StringField("Country code", validators=[validators.Length(min=2, max=2), validators.Optional()])
    billing_country_area = StringField("Country area", validators=[validators.Length(max=255)])


    def __init__(self, *args, **kwargs):
       super().__init__(*args, **kwargs)
       self.variant.choices = [(x.extra["name"], x.extra.get("localized_name", x.extra["name"])) for x in get_payment_model().list_providers()]

class SelectView(FormView):
    template_name = "form.html"
    success_url = reverse_lazy("payment-form")
    initial = {"currency": "EUR", "total": Decimal("10.0")}

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['safe_urls'] = ["", reverse("payment-form"), reverse("select-form")]
        context["mytitle"] = "Select"
        return context

    def get_form(self, form_class=None):
        formkwargs = self.get_form_kwargs()
        return SelectPaymentForm(formdata=formkwargs.get("data", None), data=formkwargs["initial"])

    def form_valid(self, form):
        payment = get_payment_model().objects.create(**form.data)
        self.request.session["paymentid"] = payment.id
        return super().form_valid(form)

    def post(self, request, *args, **kwargs):
        """
        Handles POST requests, instantiating a form instance with the passed
        POST variables and then checked for validity.
        """
        form = self.get_form()
        if form.validate():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)




from decimal import Decimal
from .logic import BasicPayment
from . import PaymentStatus, PurchasedItem, ProviderVariant
from .utils import getter_prefixed_address

__all__ = ["create_test_payment"]

PAYMENT_VARIANTS_API = {
    'default': ('web_payments_dummy.DummyProvider', {}, {}),
    'DummyProvider': ('web_payments_dummy.DummyProvider', {}, {}),
    'DirectPaymentProvider': ('web_payments_externalpayments.DirectPaymentProvider', {}, {}),
    'iban': ('web_payments_externalpayments.BankTransferProvider', {
        "iban": "GL5604449876543210",
        "bic": "DABAIE2D"}, {"localized_name": "iban"}
        ),
    }

def create_test_payment(PAYMENT_VARIANTS_API=PAYMENT_VARIANTS_API, **attributes):
    class TestPayment(BasicPayment):
        id = 523
        pk = id
        description = 'payment'
        currency = 'USD'
        status = PaymentStatus.WAITING
        message = ""
        token = "354338723"
        total = Decimal(100)
        captured_amount = Decimal("0.0")
        extra_data = ""
        variant = "undefined"
        transaction_id = None

        billing_first_name = "John"
        billing_last_name = "Smith"
        billing_address_1 = "JohnStreet 23"
        billing_address_2 = ""
        billing_city = "Neches"
        billing_postcode = "75779"
        billing_country_code = "US"
        billing_country_area = "Tennessee"
        billing_email = "example@example.com"

        get_billing_address = getter_prefixed_address("billing")
        get_shipping_address = get_billing_address

        def __init__(self, **kwargs):
            for key, val in kwargs.items():
                setattr(self, key, val)

        def get_purchased_items(self):
            return [
                PurchasedItem(
                    name='foo', quantity=10, price=Decimal('20'),
                    currency='USD', sku='bar')]

        @staticmethod
        def _list_providers_helper(item):
            _ndict = {"name": item[0]}
            _ndict.update(item[1][2])
            return ProviderVariant(item[1][0], item[1][1], _ndict)

        @classmethod
        def list_providers(cls, **kwargs):
            """ returns an iterable with ProviderVariants """
            if "name" in kwargs:
                if kwargs["name"] in PAYMENT_VARIANTS_API:
                    return [cls._list_providers_helper((kwargs["name"], PAYMENT_VARIANTS_API[kwargs["name"]]))]
                return []
            return map(cls._list_providers_helper, PAYMENT_VARIANTS_API.items())

        def get_provider_variant(self):
            variant_tup = PAYMENT_VARIANTS_API[self.variant]
            variant = ProviderVariant(variant_tup[0], variant_tup[1], {"name": self.variant})
            variant.extra.update(variant_tup[2])
            return variant

        def get_failure_url(self):
            return 'http://cancel.com'

        def get_process_url(self, extra_data=None):
            return 'http://example.com/token'

        def get_success_url(self):
            return 'http://success.com'

        def save(self, **kwargs):
            pass
    # workaround limitation in python
    for key, val in attributes.items():
        setattr(TestPayment, key, val)
    return TestPayment



"""django_test URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
try:
    from django.urls import include
    from django.urls import path_re as url
except ImportError:
    from django.conf.urls import url
    from django.conf.urls import include

from web_payments.django import urls as web_payment_urls
from .views import PaymentView, SelectView


urlpatterns = [
    url('^admin/', admin.site.urls),
    url('^payment/', include(web_payment_urls)),
    url('^form/$', PaymentView.as_view(), name="payment-form"),
    url('', SelectView.as_view(), name="select-form"),
]



from django.apps import AppConfig
from django.conf import settings

class WebPaymentsConfig(AppConfig):
    name = 'web_payments'
    def ready(self):
        from . import initialize, get_payment_model
        initialize()
        if getattr(settings, "DEBUG", False):
            get_payment_model().load_providers()



from collections import namedtuple
from ._exceptions import (RedirectNeeded, ExternalPostNeeded,
                         PaymentError, NotInitialized, NotSupported)
from .status import PaymentStatus, FraudStatus

__all__ = ["PurchasedItem", "HttpRequest"]
__all__ += ["PaymentStatus", "FraudStatus"]
__all__ += ["RedirectNeeded", "ExternalPostNeeded"]
__all__ += ["PaymentError", "NotInitialized", "NotSupported"]
__all__ += ["provider_factory"]

PurchasedItem = namedtuple('PurchasedItem',
                           ['name', 'quantity', 'price', 'currency', 'sku'])

HttpRequest = namedtuple('HttpRequest',
                         ['method', 'GET', 'POST', 'content_type'])

# extra is a dictionary
ProviderVariant = namedtuple('ProviderVariant',
                             ['path', 'config', 'extra'])

PROVIDER_CACHE = {}

def provider_factory(variant):
    '''
    Return the provider instance based on ProviderVariant
    '''
    if not variant or None in variant:
        raise ValueError('Payment variant does not exist: %s' %
                         (variant,))
    module_path, class_name = variant.path.rsplit('.', 1)
    name = variant.extra.get("name", class_name)
    if name not in PROVIDER_CACHE:
        module = __import__(
            str(module_path), globals(), locals(), [str(class_name)])
        class_ = getattr(module, class_name)
        PROVIDER_CACHE[name] = class_(**variant.config)
        PROVIDER_CACHE[name].extra = variant.extra
    return PROVIDER_CACHE[name]



from wtforms import StringField, ValidationError
from web_payments.forms import PaymentForm
from .translation import translation
_ = translation.gettext_lazy

__all__ = ["OrderIdForm", "IBANBankingForm"]

class OrderIdForm(PaymentForm):
    # only shown, return is ignored
    order = StringField(label=_("Please supply as reference"), render_kw={"readonly": True})


class IBANBankingForm(PaymentForm):
    # only shown, return is ignored
    iban = StringField(label="IBAN", render_kw={"readonly": True})
    bic = StringField(label="BIC", render_kw={"readonly": True})
    order = StringField(label=_("Please supply as reference"), render_kw={"readonly": True})




from decimal import Decimal
from urllib.parse import urlencode, urljoin

from django.db import models
from django.conf import settings

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .. import FraudStatus, PaymentStatus, ProviderVariant
from ..logic import BasicPayment
from ..utils import getter_prefixed_address
from .signals import status_changed
from . import get_base_url

__all__ = ["BasePayment", "BasePaymentWithAddress"]


class BasePayment(models.Model, BasicPayment):
    '''
    Represents a single transaction. Each instance has one or more PaymentItem.
    '''
    # overwrite variant to remove field
    #: select payment provider
    variant = models.CharField(max_length=255)
    #: Transaction status
    status = models.CharField(
        max_length=10, choices=PaymentStatus.CHOICES,
        default=PaymentStatus.WAITING)
    #: Transaction status message
    message = models.TextField(blank=True, default='')

    #: fraud status
    fraud_status = models.CharField(
        max_length=10, choices=FraudStatus.CHOICES,
        default=FraudStatus.UNKNOWN)
    #: fraud message
    fraud_message = models.TextField(blank=True, default='')

    #: for attrs pseudo dict
    extra_data = models.TextField(blank=True, default='')
    #: secret token (for get_process_url)
    token = models.CharField(max_length=36, blank=True, default='')
    #: Transaction ID (if applicable)
    transaction_id = models.CharField(max_length=255, blank=True)

    #: Currency code (may be provider-specific)
    currency = models.CharField(max_length=10)
    #: Total amount (gross)
    total = models.DecimalField(max_digits=20, decimal_places=8, default=Decimal('0.0'))
    #: captured = current captured amount
    captured_amount = models.DecimalField(
        max_digits=20, decimal_places=8, default=Decimal('0.0'))

    #: recommended for audit:
    #created = models.DateTimeField(auto_now_add=True)
    #modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def signal_status_change(self):
        status_changed.send(sender=type(self), instance=self)

    def get_process_url(self, extra_data=None):
        url = reverse('process_payment', kwargs={'token': self.token})
        url = urljoin(get_base_url(self.provider), url)
        if extra_data:
            qs = urlencode(extra_data)
            return url + '?' + qs
        return url

    @classmethod
    def list_providers(cls, **kwargs):
        """ returns an iterable with ProviderVariants """
        def _helper(item):
            t = {"name": item[0]}
            t.update(item[1][2])
            return ProviderVariant(item[1][0], item[1][1], t)
        if "name" in kwargs:
            if kwargs["name"] in settings.PAYMENT_VARIANTS_API:
                return [_helper((kwargs["name"], settings.PAYMENT_VARIANTS_API[kwargs["name"]]))]
            else:
                return []
        else:
            return map(_helper, settings.PAYMENT_VARIANTS_API.items())

    def get_provider_variant(self):
        t = settings.PAYMENT_VARIANTS_API[self.variant]
        variant = ProviderVariant(t[0], t[1], {"name": self.variant})
        variant.extra.update(t[2])
        return variant

    @classmethod
    def check_token_exists(cls, token):
        return cls._default_manager.filter(token=token).exists()

    def save(self, **kwargs):
        self.create_token()
        return models.Model.save(self, **kwargs)


class BasePaymentWithAddress(BasePayment):
    """ Has real billing address + shippingaddress alias on billing address """
    get_billing_address = getter_prefixed_address("billing")
    get_shipping_address = get_billing_address

    billing_first_name = models.CharField(max_length=256, blank=True)
    billing_last_name = models.CharField(max_length=256, blank=True)
    billing_address_1 = models.CharField(max_length=256, blank=True)
    billing_address_2 = models.CharField(max_length=256, blank=True)
    billing_email = models.EmailField(blank=True)
    billing_city = models.CharField(max_length=256, blank=True)
    billing_postcode = models.CharField(max_length=20, blank=True)
    billing_country_code = models.CharField(max_length=2, blank=True)
    billing_country_area = models.CharField(max_length=256, blank=True)

    class Meta:
        abstract = True



"""
WSGI config for django_test project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/howto/deployment/wsgi/
"""

import os
import sys
# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

sys.path.insert(0, BASE_DIR)


from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "web_payments_dummy.django_dummy.settings")

application = get_wsgi_application()



import os
import gettext
import functools
import threading


__all__ = ["web_payments_translation_path", "set_language", "get_language", "Translation", "translation"]

web_payments_translation_path = os.path.join(os.path.dirname(__file__), "locale")

class _TLocal(threading.local):
    def __init__(self, **kwargs):
        super().__init__()
        for key, val in kwargs.items():
            setattr(self, key, val)

_tlocal = _TLocal(current_language="en")

def set_language(lang):
    '''
    Set language. Default implementation
    Note: if get_language is overwritten this method should be also overwritten
    or not used
    Note: loading with django overwrites this method if not initialized
    '''
    _tlocal.current_language = lang

def get_language():
    '''
    Get language. For translations.
    Default implementation can be overwritten.
    Note: if set_language is overwritten this method should be also overwritten
    Note: loading with django overwrites this method if not initialized
    '''
    return _tlocal.current_language

class _lazy_constant(object):
    def __init__(self, func, *args, **kwargs):
        if not callable(func):
            instance_path, func_name = func
            module_path, instance_path = instance_path.rsplit(".", 1)
            module = __import__(
                str(module_path), globals(), locals(), [str(instance_path)])
            instance_ = getattr(module, instance_path)
            func = getattr(instance_, func_name)

        self.func = functools.partial(func, *args, **kwargs)
        # skip partial, provide func
        functools.update_wrapper(self, func)

    def deconstruct(self):
        selfname = ".".join([_lazy_constant.__module__, _lazy_constant.__qualname__])
        args = [(self.func.func.__self__.instance_path, self.func.func.__qualname__.split(".", 1)[1])]
        # python 3.4 compatibility
        for arg in self.func.args:
            args.append(arg)
        return (selfname, args, self.func.keywords)

    def __getattribute__(self, item):
        if item in ("func", "__dict__", "__init__", "deconstruct"):
            return super().__getattribute__(item)
        return self.func().__getattribute__(item)

    def __eq__(self, obj): return self.func().__eq__(obj)
    def __ne__(self, obj): return self.func().__ne__(obj)
    def __lt__(self, obj): return self.func().__lt__(obj)
    def __le__(self, obj): return self.func().__le__(obj)
    def __gt__(self, obj): return self.func().__gt__(obj)
    def __ge__(self, obj): return self.func().__ge__(obj)
    def __iter__(self): return self.func().__iter__()
    def __len__(self): return self.func().__len__()
    def __str__(self): return self.func()
    def __repr__(self): return "'%s'" % self.func()

class Translation(object):
    fallback = None
    domain = None
    instance_path = None
    translation_path = None
    def __init__(self, instance_path, translation_path, domain="web_payments", fallback=None):
        self.translation_path = translation_path
        self.instance_path = instance_path
        self.domain = domain
        if fallback:
            self.fallback = fallback
        else:
            self.fallback = []

    @functools.lru_cache(typed=True)
    def _trans(self, lang):
        if lang:
            return gettext.translation("web_payments", self.translation_path, [lang]+self.fallback, fallback=True)
        return gettext.translation("web_payments", self.translation_path, self.fallback, fallback=True)

    def trans(self, cur_lang=None):
        if not cur_lang:
            cur_lang = get_language()
        return self._trans(cur_lang)

    def gettext(self, msg, cur_lang=None):
        return self.trans(cur_lang).gettext(msg)

    def ngettext(self, msg, msgplural, n, cur_lang=None):
        return self.trans(cur_lang).ngettext(msg, msgplural, n)

    def gettext_lazy(self, msg, cur_lang=None):
        return _lazy_constant(self.gettext, msg, cur_lang)

    def ngettext_lazy(self, msg, msgplural, n, cur_lang=None):
        return _lazy_constant(self.ngettext, msg, msgplural, n, cur_lang)

translation = Translation("web_payments.translation.translation", web_payments_translation_path, domain="web_payments")



'''
This module is responsible for automatic processing of provider callback
data (asynchronous transaction updates).
'''
import logging

import simplejson as json
import xmltodict

from django.http import Http404, HttpResponseRedirect, HttpResponseServerError, HttpResponse
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt
from django.db.transaction import atomic
try:
    from django.urls import re_path as url
except ImportError:
    from django.conf.urls import url

from . import get_payment_model
from .. import HttpRequest, RedirectNeeded, provider_factory


def _process_data(request, payment, provider):
    try:
        # request should have some abstraction
        # redefine django request as namedtuple
        # parse most used content types correct
        if request.method == "GET":
            content = None
        elif request.content_type == "application/json":
            content = json.loads(request.body, use_decimal=True)
        elif request.content_type == "application/x-www-form-urlencoded":
            content = request.POST
        elif request.content_type in ('application/xml', 'text/xml'):
            # I cannot allow people to handle xml themselves
            # You need good security know-how to handle it
            content = xmltodict.parse(request.body)
        else:
            content = request.body
        reqparsed = HttpRequest(request.method, request.GET, content, request.content_type)
        ret = provider.process_data(payment, reqparsed)
        if ret in (True, False, None):
            status = 200
            if ret is None:
                logging.error("process_data returned None, reached end of function without returning")
                status = 500
            elif not ret:
                status = 404
            return HttpResponse(status=status)
        else:
            content, type = ret
            return HttpResponse(content, type)
    except RedirectNeeded as exc:
        return HttpResponseRedirect(exc.args[0])
    except Exception as exc:
        # for some providers this faces to the banking solution
        # log here for beeing visible
        logging.exception(exc)
        # could contain sensitive data so don't return any information
        # just log
        return HttpResponseServerError()

@csrf_exempt
@atomic
def process_data(request, token, provider=None):
    '''
    Calls process_data of an appropriate provider.

    Raises Http404 if variant does not exist.
    '''
    Payment = get_payment_model()
    payment = get_object_or_404(Payment, token=token)
    if not provider:
        try:
            provider = payment.provider
        except ValueError:
            raise Http404('No such provider')
    return _process_data(request, payment, provider)

@csrf_exempt
@atomic
def static_callback(request, variant):
    Payment = get_payment_model()
    filtered_v = Payment.list_providers(name=variant)
    if len(filtered_v) != 1:
        raise Http404('No such provider/or not unique')
    else:
        try:
            provider = provider_factory(filtered_v[0])
        except ValueError:
            raise Http404('No such provider')

    token = provider.get_token_from_request(request=request, payment=None)
    if not token:
        raise Http404('Invalid response')
    return process_data(request, token, provider)


urlpatterns = [
    url(r'^process/(?P<token>[0-9a-z]{8}-[0-9a-z]{4}-'
        '[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12})/$', process_data,
        name='process_payment'),
    url(r'^process/(?P<variant>[a-z-]+)/$', static_callback,
        name='static_process_payment')
    ]



#!/usr/bin/env python3
from setuptools import setup
from setuptools.command.test import test as TestCommand
import os
import sys

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'test_settings')

DJANGO_VERSIONS="django>=1.11"

PACKAGES = [
    'web_payments',
    'web_payments.django',
    'web_payments_externalpayments',
    'web_payments_dummy']

REQUIREMENTS = [
    'simplejson>=3.0.0',
    'wtforms<3.0.0',
    # before some people open security holes by handling xml wrong, implement it here and give them only a dict
    # required for xmltodict, if not given it uses the vulnerable implementation as fallback
    'defusedxml>=0.4.1',
    'xmltodict>=0.10.2'
]

TEST_REQUIREMENTS = [
    DJANGO_VERSIONS,
    'pytest',
    'pytest-django'
]

VERSIONING = {
    'root': '.',
    'version_scheme': 'guess-next-dev',
    'local_scheme': 'dirty-tag',
}

EXTRAS={
    'django':  [DJANGO_VERSIONS]
}

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]
    test_args = []

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
      name='web-payments-connector',
      license="MIT",
      author='Alexander Kaftan',
      author_email='devkral@web.de',
      description='Universal payment handling for Web Frameworks',
      use_scm_version=VERSIONING,
      setup_requires=['setuptools_scm'],
      url='http://github.com/devkral/web-payments-connector',
      packages=PACKAGES,
      extras_require=EXTRAS,
      include_package_data=True,
      classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules'],
      install_requires=REQUIREMENTS,
      cmdclass={
        'test': PyTest},
      tests_require=TEST_REQUIREMENTS,
      zip_safe=False)



from __future__ import unicode_literals
import os

PROJECT_ROOT = os.path.normpath(
    os.path.join(os.path.dirname(__file__), 'web_payments'))
TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': [os.path.join(PROJECT_ROOT, 'templates')]}]

SECRET_KEY = 'NOTREALLY'
PAYMENT_HOST = 'example.com'
PAYMENT_PROTOCOL = 'https'

PAYMENT_VARIANTS_API = {
    'default': ('web_payments_dummy.DummyProvider', {}, {}),
    'DummyProvider': ('web_payments_dummy.DummyProvider', {}, {}),
    'direct': ('web_payments_externalpayments.DirectPaymentProvider', {}, {}),
    'iban': ('web_payments_externalpayments.BankTransferProvider', {
        "iban": "GL5604449876543210",
        "bic": "DABAIE2D"}, {}
        ),
    }

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(os.path.dirname(__file__), 'db.sqlite3'),
    }
}

INSTALLED_APPS = ['django.contrib.sites', 'web_payments']



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
# apps clash with 'apps' of module so import as django_apps
from django.apps import apps as django_apps
from django.utils import translation as django_translation

from .. import translation


__all__ = ["get_base_url", "get_payment_model", "initialize"]

default_app_config = 'web_payments.django.apps.WebPaymentsConfig'


def get_payment_model():
    '''
    Return the Payment model that is active in this project
    '''
    try:
        app_label, model_name = settings.PAYMENT_MODEL.split('.')
    except (ValueError, AttributeError):
        raise ImproperlyConfigured('PAYMENT_MODEL must be of the form '
                                   '"app_label.model_name"')
    payment_model = django_apps.get_model(app_label, model_name)
    if payment_model is None:
        msg = (
            'PAYMENT_MODEL refers to model "%s" that has not been installed' %
            settings.PAYMENT_MODEL)
        raise ImproperlyConfigured(msg)
    return payment_model


def get_base_url(provider=None):
    """
    Returns host url according to project settings. Protocol is chosen by
    PAYMENT_PROTOCOL variable.
    If PAYMENT_HOST is not specified, get domain from Sites.
    Otherwise checks if it's callable and returns it's result. If it's not a
    callable treats it as domain.
    """
    protocol = getattr(settings, 'PAYMENT_PROTOCOL', 'https')
    PAYMENT_HOST = getattr(settings, 'PAYMENT_HOST', None)
    if not PAYMENT_HOST:
        from django.contrib.sites.models import Site
        current_site = Site.objects.get_current()
        domain = current_site.domain
    elif callable(PAYMENT_HOST):
        domain = PAYMENT_HOST(provider)
    else:
        domain = PAYMENT_HOST
    return '%s://%s' % (protocol, domain)


def initialize():
    ''' overwrites get/set language with django equivalents '''
    translation.get_language = django_translation.get_language
    translation.set_language = django_translation.activate



from uuid import uuid4
from decimal import Decimal


import simplejson as json

from . import NotSupported, FraudStatus, PaymentStatus, ProviderVariant, provider_factory

__all__ = ["BasicPayment", "BasicProvider"]


class PaymentAttributeProxy(object):

    def __init__(self, payment):
        super().__init__()
        self._payment = payment
        # replace
        #payment.__dict__["attrs"] = self

    def __getattr__(self, item):
        data = json.loads(self._payment.extra_data or '{}')
        try:
            return data[item]
        except KeyError as e:
            raise AttributeError(*e.args)

    def __setattr__(self, key, value):
        if key == '_payment':
            return super().__setattr__(key, value)
        try:
            data = json.loads(self._payment.extra_data, use_decimal=True)
        except ValueError:
            data = {}
        data[key] = value
        self._payment.extra_data = json.dumps(data, use_decimal=True)

class BasicPayment(object):
    '''
        Logic of a Payment object, basis for implementations
    '''

    #: select payment provider
    variant = NotImplemented
    #: Transaction status
    status = NotImplemented
    #: Transaction status message
    message = NotImplemented

    #: fraud status
    fraud_status = NotImplemented
    #: fraud message
    fraud_message = NotImplemented

    #: for attrs pseudo dict
    extra_data = NotImplemented
    #: secret token (for get_process_url)
    token = NotImplemented
    #: Transaction ID (if applicable)
    transaction_id = NotImplemented

    #: Currency code (may be provider-specific)
    currency = NotImplemented
    #: Total amount (gross)
    total = NotImplemented
    #: captured = current captured amount
    captured_amount = NotImplemented

    def change_status(self, status, message=''):
        '''
            Updates the Payment status and sends the status_changed signal.
        '''
        self.status = status
        self.message = message
        self.save()
        self.signal_status_change()

    def signal_status_change(self):
        '''
            Called on status change. Should send signal (see django.models for example).
            must to be overwritten to be useful
        '''
        pass

    def change_fraud_status(self, status, message='', commit=True):
        available_statuses = [choice[0] for choice in FraudStatus.CHOICES]
        if status not in available_statuses:
            raise ValueError(
                'Wrong status "%s", it should be one of: %s' % (
                    status, ', '.join(available_statuses)))
        self.fraud_status = status
        self.fraud_message = message
        if commit:
            self.save()

    def __str__(self):
        return self.variant

    def get_form(self, data=None, **kwargs):
        return self.provider.get_form(self, data=data, **kwargs)

    def get_purchased_items(self):
        return []

    def get_failure_url(self):
        '''
            url where customer should be redirected if payment had an error
        '''
        raise NotImplementedError()

    def get_success_url(self):
        '''
            url where customer should be redirected if payment was successful
        '''
        raise NotImplementedError()

    def get_process_url(self, extra_data=None):
        '''
            returns a communication url, should kept secret
            except if provider communication is with customer
        '''
        raise NotImplementedError()

    @classmethod
    def list_providers(cls, **_kwargs):
        '''
            returns an iterable with ProviderVariants
            possible keywords:
            name=<variantname>: extract variant, return list with one provider or [], required for static_callback
        '''
        raise NotImplementedError()

    def get_provider_variant(self):
        ''' return ProviderVariant for this payment object '''
        raise NotImplementedError()

    @property
    def provider(self):
        ''' returns provider object '''
        try:
            return provider_factory(self.get_provider_variant())
        except (KeyError, AttributeError) as exc:
            raise ValueError("Payment has invalid provider") from exc

    @classmethod
    def load_providers(cls):
        '''
            Load all providers in cache
            Also useful method to check if all providers are valid
        '''
        for i in cls.list_providers():
            provider_factory(i)

    def get_payment_extra(self):
        '''
            extra costs like delivery or tax (required, Decimal), defaults to zero
            Payment message, minimumage,... (not required, provider SHOULD not depend on it)
            Overwrite or extend to be useful

            universal types:
            type: what type is the transaction (official, physical, ...),
                    VALID value can be provider dependent
            message: message for customer
            minimumage: minimum age for customer
        '''
        return {
            "tax": Decimal("0"),
            "delivery": Decimal("0")
        }

    # needs to be implemented, see BasePaymentWithAddress for an example
    def get_shipping_address(self):
        ''' return shipping address '''
        raise NotImplementedError()

    # needs to be implemented, see BasePaymentWithAddress for an example
    def get_billing_address(self):
        ''' return billing address '''
        raise NotImplementedError()

    def capture(self, amount=None, final=True):
        '''
            Capture a fraction of the total amount of a payment.
            Return amount captured or None
        '''
        if self.status != PaymentStatus.PREAUTH:
            raise ValueError(
                'Only pre-authorized payments can be captured.')
        amount = self.provider.capture(self, amount, final)
        if amount:
            self.captured_amount += amount
            if final:
                self.change_status(PaymentStatus.CONFIRMED)
        return amount

    def release(self):
        ''' Annilates captured payment '''
        if self.status != PaymentStatus.PREAUTH:
            raise ValueError(
                'Only pre-authorized payments can be released.')
        self.provider.release(self)
        self.change_status(PaymentStatus.REFUNDED)

    def refund(self, amount=None):
        ''' Refund payment, return amount which was refunded '''
        if self.status != PaymentStatus.CONFIRMED:
            raise ValueError(
                'Only charged payments can be refunded.')
        if amount:
            if amount > self.captured_amount:
                raise ValueError(
                    'Refund amount can not be greater than captured amount')
        amount = self.provider.refund(self, amount)
        if amount:
            if amount > self.captured_amount:
                raise ValueError(
                    'Provider returned refund amount can not be greater than captured amount')
            self.captured_amount -= amount
            if self.captured_amount == 0 and self.status != PaymentStatus.REFUNDED:
                self.change_status(PaymentStatus.REFUNDED)
            self.save()
        return amount

    @classmethod
    def check_token_exists(cls, token):
        ''' create token for process_url '''
        return False

    def create_token(self):
        if not self.token:
            tries = set()  # Stores a set of tried values
            while True:
                token = str(uuid4())
                if token in tries and len(tries) >= 100:  # After 100 tries we are impliying an infinite loop
                    raise SystemExit('A possible infinite loop was detected')
                else:
                    if not self.check_token_exists(token):
                        self.token = token
                        break
                tries.add(token)

    @property
    def attrs(self):
        return PaymentAttributeProxy(self)

    def save(self, **kwargs):
        ''' save model implementation dependant '''
        raise NotImplementedError()


BasePaymentLogic = BasicPayment


class BasicProvider(object):
    '''
        This class defines the provider API. It should not be instantiated
        directly. Use factory instead.
    '''
    form_class = None

    def __init__(self, capture=True):
        self._capture = capture

    def get_action(self, payment):
        return ""

    def get_form(self, payment, data=None, **kwargs):
        '''
            Converts *payment* into a form
        '''
        if not self.form_class:
            raise NotSupported("No form class specified")
        return self.form_class(formdata=data, provider=self, payment=payment, **kwargs)

    def process_data(self, payment, request):
        '''
            Process callback request from a payment provider.
            Default: return 404 if somebody tries it
        '''
        return False

    def get_token_from_request(self, payment, request):
        '''
            Return payment token from provider request.
        '''
        raise NotImplementedError()

    def capture(self, payment, amount=None, final=True):
        '''
            Capture a fraction of the total amount of a payment.
            Return amount captured or None
        '''
        raise NotImplementedError()

    def release(self, payment):
        ''' Annilates captured payment '''
        raise NotImplementedError()

    def refund(self, payment, amount=None):
        ''' Refund payment, return amount which was refunded '''
        raise NotImplementedError()



from .forms import OrderIdForm, IBANBankingForm
from web_payments import RedirectNeeded, NotSupported
from web_payments.status import PaymentStatus
from web_payments.logic import BasicProvider

__all__ = ["DirectPaymentProvider", "BankTransferProvider"]

class DirectPaymentProvider(BasicProvider):
    '''
        Payment is done manually e.g. cash on delivery, voucher in restaurant.
        Payments are just a placeholder; to allow every further operation set to CONFIRMED.

        skipform:
            doesn't show form with data
        usetoken:
            if you don't verify the name; don't allow people
            supplying others order numbers (use token)
            because it is much longer it defaults to off
        prefix:
            reference: add prefix to payment id
        confirm:
            set PaymentStatus to CONFIRMED when this payment pipeline completes

    '''

    def __init__(self, skipform=True, confirm=False,
                 usetoken=False, prefix="", **kwargs):
        super(DirectPaymentProvider, self).__init__(**kwargs)
        self.skipform = skipform
        self.prefix = prefix
        self.usetoken = usetoken
        self.confirm = confirm
        if not self._capture:
            raise NotSupported(
                'Direct Payments do not support pre-authorization.')

    def get_form(self, payment, data=None):
        if not payment.id:
            payment.save()
        if not payment.transaction_id:
            if self.usetoken:
                payment.transaction_id = "{}{}-{}".format(self.prefix, payment.id, payment.token)
            else:
                payment.transaction_id = "{}{}".format(self.prefix, payment.id)
            payment.save()
        if not self.skipform:
            if not data or not data.get("order", None):
                return OrderIdForm(data={"order": payment.transaction_id},
                                   payment=payment,
                                   provider=self)
        if self.confirm:
            payment.change_status(PaymentStatus.CONFIRMED)
        else:
            payment.change_status(PaymentStatus.WAITING)
        raise RedirectNeeded(payment.get_success_url())

    def refund(self, payment, amount=None):
        if not amount:
            amount = payment.total
        payment.change_status(PaymentStatus.REFUNDED)
        return amount

class BankTransferProvider(BasicProvider):
    '''
        Banking software or human confirms transaction.
        Because there is no security problems if somebody pays for somebody else and references can not hold many characters, only the id is required.
        The form is used to show the user the data he has to send.
        iban:
            IBAN number
        bic:
            BIC number
        prefix:
            reference: add prefix to payment id
        confirm:
            set PaymentStatus to CONFIRMED when this payment pipeline completes
    '''

    def __init__(self, iban, bic, confirm=False,
                 prefix="", **kwargs):
        if len(iban) <= 10 or len(bic) <= 4:
            raise NotSupported("Wrong IBAN or BIC")
        self.iban = iban.upper()
        self.bic = bic.upper()
        self.confirm = confirm
        self.prefix = prefix
        super(BankTransferProvider, self).__init__(**kwargs)
        if not self._capture:
            raise NotSupported(
                'Advance Payment does not support pre-authorization.')

    def get_fields(self, payment):
        return {
            'iban': self.iban,
            'bic': self.bic,
            'order': payment.transaction_id
        }

    def get_form(self, payment, data=None):
        if not payment.id:
            payment.save()
        if not payment.transaction_id:
            payment.transaction_id = "{}{}".format(self.prefix, payment.id)
            payment.save()
        if not data or not data.get("order", None):
            return IBANBankingForm(data=self.get_fields(payment), payment=payment, provider=self)
        if self.confirm:
            payment.change_status(PaymentStatus.CONFIRMED)
        else:
            payment.change_status(PaymentStatus.WAITING)
        raise RedirectNeeded(payment.get_success_url())

    def refund(self, payment, amount=None):
        if not amount:
            amount = payment.total
        payment.change_status(PaymentStatus.REFUNDED)
        return amount
